package ru.dbotthepony.mc.otm.block.entity

import it.unimi.dsi.fastutil.longs.Long2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.longs.Long2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import net.minecraft.core.BlockPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.entity.player.PlayerEvent
import net.minecraftforge.event.entity.player.PlayerEvent.PlayerChangedDimensionEvent
import net.minecraftforge.event.level.ChunkWatchEvent
import net.minecraftforge.event.server.ServerStartingEvent
import net.minecraftforge.event.server.ServerStoppingEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.core.component1
import ru.dbotthepony.mc.otm.core.component2
import ru.dbotthepony.mc.otm.core.component3
import ru.dbotthepony.mc.otm.core.position
import ru.dbotthepony.mc.otm.network.BlockEntitySyncPacket
import ru.dbotthepony.mc.otm.network.FieldSynchronizer
import ru.dbotthepony.mc.otm.network.WorldNetworkChannel
import java.lang.ref.WeakReference
import java.util.LinkedList
import java.util.WeakHashMap

private data class Subscribers(
	val blockEntities: LinkedList<WeakReference<BlockEntity>> = LinkedList(),
	val players: LinkedList<ServerPlayer> = LinkedList(),
	val level: WeakReference<Level>,
	val chunkPos: Long,
	var changeset: Int = 0
) {
	fun cleanUpBlocks() {
		val listIterator = blockEntities.listIterator()

		for (block in listIterator) {
			if (block.get() == null) {
				listIterator.remove()
			}
		}
	}

	fun subscribe(blockEntity: BlockEntity) {
		for (value in blockEntities) {
			if (value.get() === blockEntity) {
				return
			}
		}

		blockEntities.add(WeakReference(blockEntity))
		changeset++
	}

	fun unsubscribe(blockEntity: BlockEntity): Boolean {
		val listIterator = blockEntities.listIterator()

		for (value in listIterator) {
			if (value.get() === blockEntity) {
				listIterator.remove()
				changeset++
				break
			}
		}

		if (players.isNotEmpty()) {
			return false
		}

		cleanUpBlocks()
		return blockEntities.isEmpty()
	}

	val isEmpty: Boolean get() {
		if (!players.isEmpty()) {
			return false
		}

		cleanUpBlocks()
		return blockEntities.isEmpty()
	}
}

abstract class SynchronizedBlockEntity(p_155228_: BlockEntityType<*>, p_155229_: BlockPos, p_155230_: BlockState) : BlockEntity(p_155228_, p_155229_, p_155230_) {
	val synchronizer = FieldSynchronizer()

	init {
		synchronizer.defaultEndpoint.markUnused()
	}

	open val synchronizationRadius: Double = 32.0

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		unsubscribe()
		_subCache = null
	}

	private var _subCache: Subscribers? = null

	private fun unsubscribe() {
		val subCache = _subCache ?: return

		if (subCache.unsubscribe(this)) {
			val level = subCache.level.get()

			if (level != null) {
				playerMap.get(level)?.remove(subCache.chunkPos)
			}
		}

		lastSubscriptionChangeset = -1
	}

	private fun subscribe(): Subscribers {
		unsubscribe()
		val subs = playerMap.computeIfAbsent(level) { Long2ObjectAVLTreeMap() }.computeIfAbsent(ChunkPos(blockPos).toLong(), Long2ObjectFunction { Subscribers(level = WeakReference(level), chunkPos = ChunkPos(blockPos).toLong()) })
		subs.subscribe(this)
		_subCache = subs
		return subs
	}

	private val subscription get() = _subCache ?: subscribe()
	private var lastSubscriptionChangeset = -1

	fun synchronizeToPlayers() {
		if (synchronizationRadius <= 0.0 || synchronizer.isEmpty) {
			return
		}

		check(level is ServerLevel) { "Invalid realm or Level is null" }

		synchronizer.observe()

		val subscription = subscription

		if (subscription.players.isEmpty() || lastSubscriptionChangeset == subscription.changeset && !synchronizer.hasChanges) {
			return
		}

		lastSubscriptionChangeset = subscription.changeset

		val (xi, yi, zi) = blockPos
		val bx = xi + 0.5
		val by = yi + 0.5
		val bz = zi + 0.5

		val double = synchronizationRadius * synchronizationRadius

		subscription.players.stream().filter {
			if (!it.isAlive) {
				return@filter false
			}

			val (x, y, z) = it.position

			(x - bx) * (x - bx) +
			(y - by) * (y - by) +
			(z - bz) * (z - bz) <= double
		}.forEach {
			val payload = synchronizer.computeEndpointFor(it).collectNetworkPayload()

			if (payload != null) {
				WorldNetworkChannel.send(it, BlockEntitySyncPacket(blockPos, payload.array, payload.length))
			}
		}
	}

	/**
	 * Why track player-tracked chunks?
	 *
	 * because minecraft itself doesn't track them, well,
	 * in "section partitioning" way.
	 *
	 * just look at [net.minecraft.server.level.PlayerMap]
	 *
	 * the [net.minecraft.server.level.PlayerMap.getPlayers] straight ignores chunk position
	 * and just fetches all players
	 *
	 * even if they did not ignore that argument, you still have to fetch player *list* though
	 * [net.minecraft.server.level.ChunkMap.getPlayers], which is not something we want
	 */
	companion object {
		private val playerMap = WeakHashMap<Level, Long2ObjectAVLTreeMap<Subscribers>>()

		fun onServerStopping(event: ServerStoppingEvent) {
			playerMap.clear()
		}

		fun onServerStarting(event: ServerStartingEvent) {
			playerMap.clear()
		}

		fun onWatch(event: ChunkWatchEvent.Watch) {
			playerMap.computeIfAbsent(event.level) { Long2ObjectAVLTreeMap() }.computeIfAbsent(event.pos.toLong(), Long2ObjectFunction { Subscribers(level = WeakReference(event.level), chunkPos = event.pos.toLong()) }).let {
				val (blocks, players) = it

				if (event.player !in players) {
					players.add(event.player)
					it.changeset++
				}
			}
		}

		fun onForget(event: ChunkWatchEvent.UnWatch) {
			val subs = playerMap.get(event.level)?.get(event.pos.toLong()) ?: return

			if (subs.players.remove(event.player)) {
				if (subs.isEmpty) {
					playerMap.get(event.level)?.remove(event.pos.toLong())
				} else {
					subs.changeset++
				}
			}
		}

		fun playerDisconnected(event: PlayerEvent.PlayerLoggedOutEvent) {
			for (tree in playerMap.values) {
				val iterator = tree.iterator()

				for (entry in iterator) {
					val listIterator = entry.value.players.listIterator()

					for (value in listIterator) {
						if (value === event.entity) {
							listIterator.remove()
							entry.value.changeset++
						}
					}

					if (entry.value.isEmpty) {
						iterator.remove()
					}
				}
			}
		}
	}
}
