package ru.dbotthepony.mc.otm.client.render

sealed interface IUVCoords {
	val u0: Float
	val v0: Float
	val u1: Float
	val v1: Float
	operator fun get(index: Int): Float
}

private val mapImmutable = ArrayList<(input: UVCoords) -> Float>()
	.also { it.add { it.u0 } }
	.also { it.add { it.v0 } }
	.also { it.add { it.u1 } }
	.also { it.add { it.v1 } }
	.toTypedArray()

private val mapMutable = ArrayList<(input: MutableUVCoords) -> Float>()
	.also { it.add { it.u0 } }
	.also { it.add { it.v0 } }
	.also { it.add { it.u1 } }
	.also { it.add { it.v1 } }
	.toTypedArray()

private val mapSetMutable = ArrayList<(input: MutableUVCoords, value: Float) -> Unit>()
	.also { it.add { input, value -> input.u0 = value } }
	.also { it.add { input, value -> input.v0 = value } }
	.also { it.add { input, value -> input.u1 = value } }
	.also { it.add { input, value -> input.v1 = value } }
	.toTypedArray()

data class UVCoords(
	override val u0: Float,
	override val v0: Float,
	override val u1: Float,
	override val v1: Float,
) : IUVCoords {
	override fun get(index: Int): Float {
		return mapImmutable[index].invoke(this)
	}
}

data class MutableUVCoords(
	override var u0: Float,
	override var v0: Float,
	override var u1: Float,
	override var v1: Float,
) : IUVCoords {
	override fun get(index: Int): Float {
		return mapMutable[index].invoke(this)
	}

	operator fun set(index: Int, value: Float) {
		mapSetMutable[index].invoke(this, value)
	}
}

private val pickers = ArrayList<(u0: Float,
								 v0: Float,
								 u1: Float,
								 v1: Float) -> Float>()
	.also { it.add { u0, _, _, _ -> u0 } }
	.also { it.add { _, v0, _, _ -> v0 } }
	.also { it.add { _, _, u1, _ -> u1 } }
	.also { it.add { _, _, _, v1 -> v1 } }
	.toTypedArray()

enum class UVWindingOrder(
	val u0: Int,
	val v0: Int,
	val u1: Int,
	val v1: Int,
) {
	NORMAL(0, 1, 2, 3), // normal operation
	U0_V0_U1_V1(0, 1, 2, 3), // normal operation
	U0_V1_U1_V0(0, 3, 2, 1), // mirror y
	FLIP(0, 3, 2, 1), // mirror y
	U1_V0_U0_V1(2, 1, 0, 3), // mirror x
	FLOP(2, 1, 0, 3), // mirror x
	U1_V1_U0_V0(2, 3, 0, 1), // mirror both
	FLIP_FLOP(2, 3, 0, 1); // mirror both

	val isIdentity: Boolean = u0 == 0 && v0 == 1 && u1 == 2 && v1 == 3

	operator fun component1() = u0
	operator fun component2() = v0
	operator fun component3() = u1
	operator fun component4() = v1

	val u0Picker = pickers[u0]
	val v0Picker = pickers[v0]
	val u1Picker = pickers[u1]
	val v1Picker = pickers[v1]

	private val buffer = FloatArray(4)

	fun translate(
		u0: Float,
		v0: Float,
		u1: Float,
		v1: Float,
	): UVCoords {
		synchronized(buffer) {
			buffer[0] = u0
			buffer[1] = v0
			buffer[2] = u1
			buffer[3] = v1

			return UVCoords(buffer[this.u0], buffer[this.v0], buffer[this.u1], buffer[this.v1])
		}
	}

	fun translate(
		input: UVCoords
	): UVCoords {
		synchronized(buffer) {
			buffer[0] = input.u0
			buffer[1] = input.v0
			buffer[2] = input.u1
			buffer[3] = input.v1

			return UVCoords(buffer[this.u0], buffer[this.v0], buffer[this.u1], buffer[this.v1])
		}
	}

	fun translate(
		input: MutableUVCoords
	): MutableUVCoords {
		synchronized(buffer) {
			buffer[0] = input.u0
			buffer[1] = input.v0
			buffer[2] = input.u1
			buffer[3] = input.v1

			input[0] = buffer[this.u0]
			input[1] = buffer[this.v0]
			input[2] = buffer[this.u1]
			input[3] = buffer[this.v1]

			return input
		}
	}
}
