package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.Streams
import com.mojang.blaze3d.platform.InputConstants
import net.minecraft.ChatFormatting
import net.minecraft.client.Minecraft
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.ItemLike
import net.minecraftforge.event.entity.player.ItemTooltipEvent
import net.minecraftforge.event.entity.player.PlayerEvent
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import org.lwjgl.glfw.GLFW
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.drive.IMatteryDrive
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.network.MatteryPacket
import ru.dbotthepony.mc.otm.network.RegistryNetworkChannel
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper
import java.math.BigInteger
import java.util.function.Supplier
import java.util.stream.Stream
import kotlin.math.pow

internal var building = false

interface IMatterItem {
	fun getMatterValue(stack: ItemStack): MatterTuple?
	fun hasMatterValue(stack: ItemStack) = getMatterValue(stack) != null
	fun canDecompose(stack: ItemStack): Boolean
}

private val rootEntries = HashMap<Item, MatterTuple>()
private val derivedEntries = HashMap<Item, MatterTuple>()

/**
 * Nothing prevents you from writing to this map.
 * This derived class is designed solely to network YOUR changes to players online in real-time.
 *
 * If you use MapEntry#setValue, it is your fault this map go out of sync with clients.
 */
@JvmField
val rootEntriesAccess = object : MutableMap<Item, MatterTuple> by rootEntries {
	override fun put(key: Item, value: MatterTuple): MatterTuple? {
		RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), RegistryPacketUpdate(MatterRegistryType.ROOT, key, value))
		return rootEntries.put(key, value)
	}

	override fun putAll(from: Map<out Item, MatterTuple>) {
		rootEntries.putAll(from)

		if (!building) {
			val distr = PacketDistributor.ALL.noArg()

			for ((key, value) in from) {
				RegistryNetworkChannel.send(distr, RegistryPacketUpdate(MatterRegistryType.ROOT, key, value))
			}
		}
	}

	override fun remove(key: Item): MatterTuple? {
		val value = rootEntries.remove(key)

		if (!building && value != null)
			RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), RegistryPacketRemove(MatterRegistryType.ROOT, key))

		return value
	}

	override fun clear() {
		rootEntries.clear()

		if (!building && NULLABLE_MINECRAFT_SERVER != null) {
			RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), RegistryPacketClear.ROOT)
		}
	}
}

/**
 * Nothing prevents you from writing to this map.
 * This derived class is designed solely to network YOUR changes to players online in real-time.
 *
 * If you use MapEntry#setValue, it is your fault this map go out of sync with clients.
 */
@JvmField
val derivedEntriesAccess = object : MutableMap<Item, MatterTuple> by derivedEntries {
	override fun put(key: Item, value: MatterTuple): MatterTuple? {
		RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), RegistryPacketUpdate(MatterRegistryType.DERIVED, key, value))
		return derivedEntries.put(key, value)
	}

	override fun putAll(from: Map<out Item, MatterTuple>) {
		derivedEntries.putAll(from)

		if (!building) {
			val distr = PacketDistributor.ALL.noArg()

			for ((key, value) in from) {
				RegistryNetworkChannel.send(distr, RegistryPacketUpdate(MatterRegistryType.DERIVED, key, value))
			}
		}
	}

	override fun remove(key: Item): MatterTuple? {
		val value = derivedEntries.remove(key)

		if (!building && value != null)
			RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), RegistryPacketRemove(MatterRegistryType.DERIVED, key))

		return value
	}

	override fun clear() {
		derivedEntries.clear()

		if (!building && NULLABLE_MINECRAFT_SERVER != null) {
			RegistryNetworkChannel.send(PacketDistributor.ALL.noArg(), RegistryPacketClear.DERIVED)
		}
	}
}

val matterRegistryStream: Stream<Map.Entry<Item, MatterTuple>> get() {
	return Streams.concat(rootEntries.entries.stream(), derivedEntries.entries.stream())
}

fun getMatterValue(item: Item): MatterTuple {
	if (item is IMatterItem)
		return item.getMatterValue(ItemStack.EMPTY) ?: MatterTuple.ZERO

	return rootEntries[item] ?: derivedEntries[item] ?: MatterTuple.ZERO
}

fun getMatterValueNullable(item: Item): MatterTuple? {
	if (item is IMatterItem)
		return item.getMatterValue(ItemStack.EMPTY)

	return rootEntries[item] ?: derivedEntries[item]
}

fun hasMatterValue(item: Item): Boolean {
	if (item is IMatterItem)
		return item.hasMatterValue(ItemStack.EMPTY)

	return !getMatterValue(item).isZero
}

fun hasMatterValue(stack: ItemStack): Boolean {
	val item = stack.item

	if (item is IMatterItem)
		return item.hasMatterValue(stack)

	return !getMatterValue(stack).isZero
}

fun canDecompose(item: Item): Boolean {
	if (item is IMatterItem)
		return item.canDecompose(ItemStack.EMPTY)

	return hasMatterValue(item)
}

fun canDecompose(stack: ItemStack): Boolean {
	val item = stack.item

	if (item is IMatterItem)
		return item.canDecompose(stack)

	return canDecompose(stack.item) &&
		(stack.getCapability(MatteryCapability.MATTER).orNull()?.storedMatter ?: ImpreciseFraction.ZERO).isZero &&
		(stack.getCapability(MatteryCapability.DRIVE).orNull()?.storedCount ?: BigInteger.ZERO).isZero
}

private const val MAX_NESTING = 100

private fun getMatterValue(stack: ItemStack, level: Int): MatterTuple {
	if (level >= MAX_NESTING) {
		return MatterTuple.ZERO
	}

	var matter = getMatterValue(stack.item)

	if (stack.item is IMatterItem)
		matter = (stack.item as IMatterItem).getMatterValue(stack) ?: MatterTuple.ZERO

	if (!matter.isZero && stack.isDamageableItem) {
		val severity = 1.0 - stack.damageValue.toDouble() / stack.maxDamage.toDouble()
		matter = MatterTuple(matter.value * severity, matter.complexity * (0.5 + severity / 2))
	}

	val matterCap = stack.getCapability(MatteryCapability.MATTER).orNull()

	if (matterCap != null) {
		matter = MatterTuple(matter.value + matterCap.storedMatter, matter.complexity)
	}

	val drive = stack.getCapability(MatteryCapability.DRIVE).orNull()

	if (drive != null && drive.storageType === OverdriveThatMatters.INSTANCE.ITEM_STORAGE()) {
		for (item in (drive as IMatteryDrive<ItemStackWrapper>).stacks) {
			val tuple = getMatterValue(item.stack.stack, level + 1)

			if (!tuple.isZero) {
				matter += tuple
			}
		}
	}

	return matter
}

/**
 * Возвращает массу предмета БЕЗ учёта размера стопки
 */
fun getMatterValue(stack: ItemStack): MatterTuple {
	return getMatterValue(stack, 0)
}

fun getResearchAdvance(complexity: Double): Double {
	if (complexity <= 0.0) {
		return 0.0
	}

	return (1.0 / (complexity.pow(1.0 / 3.0) * 12.0).coerceAtLeast(0.01)).coerceAtMost(0.2).coerceAtLeast(0.025)
}

fun getResearchAdvance(stack: ItemStack): Double {
	return getResearchAdvance(getMatterValue(stack).complexity)
}

fun getResearchAdvance(stack: ItemLike): Double {
	return getResearchAdvance(getMatterValue(stack.asItem()).complexity)
}

fun onPlayerJoin(event: PlayerEvent.PlayerLoggedInEvent) {
	if (NULLABLE_MINECRAFT_SERVER != null) {
		RegistryNetworkChannel.send(PacketDistributor.PLAYER.with { event.entity as ServerPlayer }, RegistryPacketFullUpdate(MatterRegistryType.ROOT, rootEntries))
		RegistryNetworkChannel.send(PacketDistributor.PLAYER.with { event.entity as ServerPlayer }, RegistryPacketFullUpdate(MatterRegistryType.DERIVED, derivedEntries))
	}
}

fun tooltipEvent(event: ItemTooltipEvent) {
	val window = Minecraft.getInstance().window.window

	if (InputConstants.isKeyDown(window, GLFW.GLFW_KEY_LEFT_SHIFT) || InputConstants.isKeyDown(window, GLFW.GLFW_KEY_RIGHT_SHIFT)) {
		val matter = getMatterValue(event.itemStack)

		if (!matter.isZero) {
			if (matter.complexity >= 1.0E-3) {
				event.toolTip.add(
					TranslatableComponent("otm.gui.matter.format_and_complexity",
						matter.value.formatMatterFull(),
						"%.3f".format(matter.complexity)
					).withStyle(ChatFormatting.AQUA))
			} else {
				event.toolTip.add(matter.value.formatMatterFull().withStyle(ChatFormatting.AQUA))
			}
		}
	}
}

data class MatterTuple(val value: ImpreciseFraction, val complexity: Double): Comparable<MatterTuple> {
	operator fun plus(other: MatterTuple) = MatterTuple(value + other.value, complexity + other.complexity)
	operator fun minus(other: MatterTuple) = MatterTuple(value - other.value, complexity - other.complexity)

	override operator fun compareTo(other: MatterTuple): Int {
		if (compareValues(other) != 0) {
			return compareComplexity(other)
		}

		return 0
	}

	fun compareValues(other: MatterTuple) = value.compareTo(other.value)
	fun compareComplexity(other: MatterTuple): Int {
		if (complexity == other.complexity)
			return 0

		return if (complexity < other.complexity) -1 else 1
	}

	fun write(buff: FriendlyByteBuf) {
		buff.writeImpreciseFraction(value)
		buff.writeDouble(complexity)
	}

	val isZero get() = value.isZero
	val isZeroComplexity get() = complexity <= 0.0

	companion object {
		@JvmField val ZERO = MatterTuple(ImpreciseFraction.ZERO, 0.0)

		fun read(buff: FriendlyByteBuf): MatterTuple {
			return MatterTuple(buff.readImpreciseFraction(), buff.readDouble())
		}
	}
}

internal enum class MatterRegistryType {ROOT, DERIVED}

internal class RegistryPacketUpdate(val type: MatterRegistryType, val item: Item, val tuple: MatterTuple) : MatteryPacket {
	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true

		when (type) {
			MatterRegistryType.ROOT -> {
				synchronized(rootEntries) {
					rootEntries[item] = tuple
				}
			}

			MatterRegistryType.DERIVED -> {
				synchronized(derivedEntries) {
					derivedEntries[item] = tuple
				}
			}
		}
	}

	override fun write(buff: FriendlyByteBuf) {
		buff.writeEnum(type)
		buff.writeRegistryId(item)
		tuple.write(buff)
	}

	companion object {
		fun read(buff: FriendlyByteBuf): RegistryPacketUpdate {
			return RegistryPacketUpdate(buff.readEnum(MatterRegistryType::class.java), buff.readRegistryIdSafe(Item::class.java), MatterTuple.read(buff))
		}
	}
}

internal class RegistryPacketRemove(val type: MatterRegistryType, val item: Item) : MatteryPacket {
	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true

		when (type) {
			MatterRegistryType.ROOT -> {
				synchronized(rootEntries) {
					rootEntries.remove(item)
				}
			}

			MatterRegistryType.DERIVED -> {
				synchronized(derivedEntries) {
					derivedEntries.remove(item)
				}
			}
		}
	}

	override fun write(buff: FriendlyByteBuf) {
		buff.writeEnum(type)
		buff.writeRegistryId(item)
	}

	companion object {
		fun read(buff: FriendlyByteBuf): RegistryPacketRemove {
			return RegistryPacketRemove(buff.readEnum(MatterRegistryType::class.java), buff.readRegistryIdSafe(Item::class.java))
		}
	}
}

internal class RegistryPacketFullUpdate(val type: MatterRegistryType, val map: Map<Item, MatterTuple>) : MatteryPacket {
	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true

		when (type) {
			MatterRegistryType.ROOT -> {
				synchronized(rootEntries) {
					rootEntries.clear()
					rootEntries.putAll(map)
				}
			}

			MatterRegistryType.DERIVED -> {
				synchronized(derivedEntries) {
					derivedEntries.clear()
					derivedEntries.putAll(map)
				}
			}
		}
	}

	override fun write(buff: FriendlyByteBuf) {
		buff.writeEnum(type)
		buff.writeInt(map.size)

		for ((key, value) in map) {
			buff.writeRegistryId(key)
			value.write(buff)
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): RegistryPacketFullUpdate {
			val map = HashMap<Item, MatterTuple>()
			val type = buff.readEnum(MatterRegistryType::class.java)

			val size = buff.readInt()

			for (i in 0 until size) {
				map[buff.readRegistryId()] = MatterTuple.read(buff)
			}

			return RegistryPacketFullUpdate(type, map)
		}
	}
}

internal enum class RegistryPacketClear(val type: MatterRegistryType) : MatteryPacket {
	ROOT(MatterRegistryType.ROOT),
	DERIVED(MatterRegistryType.DERIVED);

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true

		when (type) {
			MatterRegistryType.ROOT -> {
				synchronized(rootEntries) {
					rootEntries.clear()
				}
			}

			MatterRegistryType.DERIVED -> {
				synchronized(derivedEntries) {
					derivedEntries.clear()
				}
			}
		}
	}

	override fun write(buff: FriendlyByteBuf) {
		buff.writeEnum(type)
	}

	companion object {
		fun read(buff: FriendlyByteBuf): RegistryPacketClear {
			return buff.readEnum(RegistryPacketClear::class.java)
		}
	}
}
