package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.*
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.lang.ref.WeakReference
import java.math.BigDecimal
import java.util.*
import java.util.function.Consumer
import kotlin.ConcurrentModificationException
import kotlin.collections.ArrayList
import kotlin.collections.HashMap
import kotlin.properties.ReadOnlyProperty
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KMutableProperty0
import kotlin.reflect.KProperty

interface FieldAccess<V> {
	fun read(): V
	fun write(value: V)
}

fun interface FieldGetter<V> {
	fun invoke(field: FieldAccess<V>): V
}

fun interface FieldSetter<V> {
	fun invoke(value: V, field: FieldAccess<V>, setByRemote: Boolean)
}

sealed interface IField<V> : ReadOnlyProperty<Any, V> {
	fun observe()
	fun markDirty()
	fun markDirty(endpoint: FieldSynchronizer.Endpoint)
	val value: V

	fun write(stream: DataOutputStream, endpoint: FieldSynchronizer.Endpoint)
	fun read(stream: DataInputStream)

	override fun getValue(thisRef: Any, property: KProperty<*>): V {
		return value
	}
}

sealed interface IMutableField<V> : IField<V>, ReadWriteProperty<Any, V> {
	override var value: V

	override fun setValue(thisRef: Any, property: KProperty<*>, value: V) {
		this.value = value
	}

	override fun getValue(thisRef: Any, property: KProperty<*>): V {
		return super.getValue(thisRef, property)
	}
}

data class MapChangeset<out K, out V>(
	val action: MapAction,
	val key: K?,
	val value: V?
)

enum class MapAction {
	CLEAR, ADD, REMOVE
}

class FieldSynchronizer {
	private val fields = ArrayList<IField<*>>()
	private val observers = LinkedList<IField<*>>()

	val isEmpty: Boolean get() = fields.isEmpty()
	val isNotEmpty: Boolean get() = fields.isNotEmpty()

	var hasChanges: Boolean = false
		private set

	fun byte(
		value: Byte = 0,
		getter: FieldGetter<Byte>? = null,
		setter: FieldSetter<Byte>? = null,
	): Field<Byte> {
		return Field(value, ByteValueCodec, getter, setter)
	}

	fun bool(
		value: Boolean = false,
		getter: FieldGetter<Boolean>? = null,
		setter: FieldSetter<Boolean>? = null,
	): Field<Boolean> {
		return Field(value, BooleanValueCodec, getter, setter)
	}

	fun short(
		value: Short = 0,
		getter: FieldGetter<Short>? = null,
		setter: FieldSetter<Short>? = null,
	): Field<Short> {
		return Field(value, ShortValueCodec, getter, setter)
	}

	fun long(
		value: Long = 0L,
		getter: FieldGetter<Long>? = null,
		setter: FieldSetter<Long>? = null,
	): Field<Long> {
		return Field(value, VarLongValueCodec, getter, setter)
	}

	fun fixedLong(
		value: Long = 0L,
		getter: FieldGetter<Long>? = null,
		setter: FieldSetter<Long>? = null,
	): Field<Long> {
		return Field(value, LongValueCodec, getter, setter)
	}

	fun float(
		value: Float = 0f,
		getter: FieldGetter<Float>? = null,
		setter: FieldSetter<Float>? = null,
	): Field<Float> {
		return Field(value, FloatValueCodec, getter, setter)
	}

	fun double(
		value: Double = 0.0,
		getter: FieldGetter<Double>? = null,
		setter: FieldSetter<Double>? = null,
	): Field<Double> {
		return Field(value, DoubleValueCodec, getter, setter)
	}

	fun uuid(
		value: UUID = UUID(0L, 0L),
		getter: FieldGetter<UUID>? = null,
		setter: FieldSetter<UUID>? = null,
	): Field<UUID> {
		return Field(value, UUIDValueCodec, getter, setter)
	}

	fun int(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,
	): Field<Int> {
		return Field(value, VarIntValueCodec, getter, setter)
	}

	fun fixedInt(
		value: Int = 0,
		getter: FieldGetter<Int>? = null,
		setter: FieldSetter<Int>? = null,
	): Field<Int> {
		return Field(value, IntValueCodec, getter, setter)
	}

	fun fraction(
		value: ImpreciseFraction = ImpreciseFraction.ZERO,
		getter: FieldGetter<ImpreciseFraction>? = null,
		setter: FieldSetter<ImpreciseFraction>? = null,
	): Field<ImpreciseFraction> {
		return Field(value, ImpreciseFractionValueCodec, getter, setter)
	}

	fun bigDecimal(
		value: BigDecimal = BigDecimal.ZERO,
		getter: FieldGetter<BigDecimal>? = null,
		setter: FieldSetter<BigDecimal>? = null,
	): Field<BigDecimal> {
		return Field(value, BigDecimalValueCodec, getter, setter)
	}

	fun <T : Enum<T>> enum(
		type: Class<T>,
		value: T = type.enumConstants[0],
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
	): Field<T> {
		return Field(value, EnumValueCodec(type), getter, setter)
	}

	fun <T : Enum<T>> enum(
		value: T,
		getter: FieldGetter<T>? = null,
		setter: FieldSetter<T>? = null,
	): Field<T> {
		return Field(value, EnumValueCodec(value::class.java), getter, setter)
	}

	fun item(
		value: ItemStack = ItemStack.EMPTY,
		getter: FieldGetter<ItemStack>? = null,
		setter: FieldSetter<ItemStack>? = null,
	): Field<ItemStack> {
		return Field(value, ItemStackValueCodec, getter, setter, isObserver = true)
	}

	fun item(
		delegate: KMutableProperty0<ItemStack>
	): ObservedField<ItemStack> {
		return ObservedField(delegate, ItemStackValueCodec)
	}

	fun <K, V> map(
		keyCodec: IStreamCodec<K>,
		valueCodec: IStreamCodec<V>,
		callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null,
		backingMap: MutableMap<K, V> = HashMap(),
		observingBackingMap: MutableMap<K, V>? = null,
	): Map<K, V> {
		return Map(
			keyCodec = keyCodec,
			valueCodec = valueCodec,
			callback = callback,
			backingMap = backingMap,
			observingBackingMap = observingBackingMap
		)
	}

	private val endpoints = LinkedList<WeakReference<Endpoint>>()
	val defaultEndpoint = Endpoint()

	private var lastEndpointsCleanup = System.nanoTime()

	private fun notifyEndpoints(dirtyField: IField<*>) {
		hasChanges = true

		forEachEndpoint {
			it.addDirtyField(dirtyField)
		}
	}

	private inline fun forEachEndpoint(execute: (Endpoint) -> Unit) {
		lastEndpointsCleanup = System.nanoTime()

		synchronized(endpoints) {
			val iterator = endpoints.listIterator()

			for (value in iterator) {
				val endpoint = value.get()

				if (endpoint == null) {
					iterator.remove()
				} else {
					execute.invoke(endpoint)
				}
			}
		}
	}

	inner class Endpoint {
		init {
			endpoints.addLast(WeakReference(this))

			if (System.nanoTime() - lastEndpointsCleanup >= 60_000_000_000) {
				lastEndpointsCleanup = System.nanoTime()

				synchronized(endpoints) {
					val iterator = endpoints.listIterator()

					for (value in iterator) {
						if (value.get() == null) {
							iterator.remove()
						}
					}
				}
			}
		}

		private val dirtyFields = LinkedList<IField<*>>()
		private val mapBacklogs = Reference2ObjectOpenHashMap<Map<*, *>, LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

		var unused: Boolean = false
			private set

		fun markUnused() {
			require(this === defaultEndpoint) { "This is not a default endpoint" }
			if (unused) return
			unused = true
			mapBacklogs.clear()
			dirtyFields.clear()

			val iterator = endpoints.listIterator()

			for (value in iterator) {
				if (value.get() === this) {
					iterator.remove()
				}
			}
		}

		init {
			for (field in fields) {
				field.markDirty(this)
			}
		}

		internal fun addDirtyField(field: IField<*>) {
			if (unused) {
				return
			}

			if (field !in dirtyFields) {
				dirtyFields.addLast(field)
			}
		}

		internal fun <K, V> getMapBacklog(map: Map<K, V>): LinkedList<Pair<Any?, (DataOutputStream) -> Unit>> {
			if (unused) {
				return LinkedList()
			}

			return mapBacklogs.computeIfAbsent(map, Reference2ObjectFunction {
				LinkedList()
			})
		}

		fun collectNetworkPayload(): FastByteArrayOutputStream? {
			if (unused || dirtyFields.isEmpty()) {
				return null
			}

			val stream = FastByteArrayOutputStream()
			val dataStream = DataOutputStream(stream)

			for (field in dirtyFields) {
				field.write(dataStream, this)
			}

			dirtyFields.clear()
			dataStream.write(0)

			return stream
		}
	}

	private val boundEndpoints = WeakHashMap<Any, Endpoint>()

	fun computeEndpointFor(obj: Any): Endpoint {
		return boundEndpoints.computeIfAbsent(obj) { Endpoint() }
	}

	fun endpointFor(obj: Any): Endpoint? {
		return boundEndpoints[obj]
	}

	inner class Field<V>(
		private var field: V,
		private val codec: IStreamCodec<V>,
		private val getter: FieldGetter<V>? = null,
		private val setter: FieldSetter<V>? = null,
		isObserver: Boolean = false,
	) : IMutableField<V> {
		private var remote: V = codec.copy(field)

		val id = fields.size + 1

		init {
			fields.add(this)

			if (isObserver) {
				observers.add(this)
			}
		}

		private var isDirty = false

		private val access = object : FieldAccess<V> {
			override fun read(): V {
				return field
			}

			override fun write(value: V) {
				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this@Field.field = value
			}
		}

		override fun observe() {
			if (!isDirty && !codec.compare(remote, field)) {
				notifyEndpoints(this@Field)
				isDirty = true
			}
		}

		override var value: V
			get() {
				val getter = this.getter

				if (getter != null) {
					return getter.invoke(access)
				}

				return this.field
			}
			set(value) {
				val setter = this.setter

				if (setter != null) {
					setter.invoke(value, access, false)
					return
				}

				if (this.field == value) {
					return
				}

				if (!isDirty && !codec.compare(remote, value)) {
					notifyEndpoints(this@Field)
					isDirty = true
				}

				this.field = value
			}

		override fun markDirty() {
			if (!isDirty) {
				notifyEndpoints(this@Field)
				isDirty = true
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			stream.write(id)
			codec.write(stream, field)
			isDirty = false
			remote = codec.copy(field)
		}

		override fun read(stream: DataInputStream) {
			val value = codec.read(stream)
			val setter = this.setter

			if (setter != null) {
				setter.invoke(value, access, true)
				return
			}

			this.field = value
		}
	}

	inner class ObservedField<V> : IMutableField<V> {
		private val codec: StreamCodec<V>
		private val getter: () -> V
		private val setter: (V) -> Unit

		private var remote: V

		override var value: V
			get() = getter.invoke()
			set(value) { setter.invoke(value) }

		constructor(field: KMutableProperty0<V>, dispatcher: StreamCodec<V>) {
			this.codec = dispatcher
			getter = field::get
			setter = field::set
			remote = dispatcher.copy(value)
		}

		constructor(getter: () -> V, setter: (V) -> Unit, dispatcher: StreamCodec<V>) {
			this.codec = dispatcher
			this.getter = getter
			this.setter = setter
			remote = dispatcher.copy(value)
		}

		private var isDirty = false

		val id = fields.size + 1

		init {
			fields.add(this)
			observers.add(this)
		}

		override fun observe() {
			if (!isDirty && !codec.compare(remote, value)) {
				notifyEndpoints(this)
				isDirty = true
			}
		}

		override fun markDirty() {
			if (!isDirty) {
				notifyEndpoints(this)
				isDirty = true
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			endpoint.addDirtyField(this)
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			stream.write(id)
			val value = value
			codec.write(stream, value)
			isDirty = false
			remote = codec.copy(value)
		}

		override fun read(stream: DataInputStream) {
			this.value = codec.read(stream)
		}
	}

	companion object {
		private val ClearBacklogEntry = { stream: DataOutputStream -> stream.write(MapAction.CLEAR.ordinal + 1) }
		private val MapActionList = MapAction.values()
		private val ClearMapChangeset = MapChangeset(MapAction.CLEAR, null, null)
	}

	inner class Map<K, V>(
		private val keyCodec: IStreamCodec<K>,
		private val valueCodec: IStreamCodec<V>,
		private val backingMap: MutableMap<K, V>,
		private val observingBackingMap: MutableMap<K, V>? = null,
		private val callback: ((changes: Collection<MapChangeset<K, V>>) -> Unit)? = null
	) : IField<MutableMap<K, V>> {
		private var isDirty = false
		private var sentAllValues = false
		private var isRemote = false

		val id = fields.size + 1

		init {
			fields.add(this)

			if (observingBackingMap != null)
				observers.add(this)
		}

		private fun pushBacklog(key: Any?, value: (DataOutputStream) -> Unit) {
			forEachEndpoint {
				val list = it.getMapBacklog(this)
				val iterator = list.listIterator()

				for (pair in iterator) {
					if (pair.first == key) {
						iterator.remove()
					}
				}

				list.addLast(key to value)
			}
		}

		private fun clearBacklog() {
			forEachEndpoint {
				it.getMapBacklog(this).clear()
			}
		}

		override fun observe() {
			if (isRemote) {
				return
			}

			val observingBackingMap = observingBackingMap

			if (observingBackingMap != null) {
				for ((key, value) in backingMap) {
					val remoteValue = observingBackingMap[key] ?: throw ConcurrentModificationException("Backing map of $this was modified externally, or $value missed a modification")

					if (!valueCodec.compare(value, remoteValue)) {
						val valueCopy = valueCodec.copy(value)

						pushBacklog(key) {
							it.write(MapAction.ADD.ordinal + 1)
							keyCodec.write(it, key)
							valueCodec.write(it, valueCopy)
						}

						observingBackingMap[key] = valueCopy

						if (!isDirty) {
							notifyEndpoints(this)
							isDirty = true
						}
					}
				}
			}
		}

		override fun markDirty() {
			if (isRemote) {
				return
			}

			if (endpoints.isEmpty()) {
				val observingBackingMap = observingBackingMap ?: return

				for ((key, value) in backingMap)
					observingBackingMap[key] = valueCodec.copy(value)

				return
			}

			isDirty = true
			val backlogs = LinkedList<LinkedList<Pair<Any?, (DataOutputStream) -> Unit>>>()

			forEachEndpoint {
				it.addDirtyField(this)
				val value = it.getMapBacklog(this)
				backlogs.add(value)
				value.clear()
				value.add(null to ClearBacklogEntry)
			}

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				val action = { it: DataOutputStream ->
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				for (backlog in backlogs) {
					backlog.add(key to action)
				}

				observingBackingMap?.put(key, valueCopy)
			}
		}

		override fun markDirty(endpoint: Endpoint) {
			if (isRemote) {
				return
			}

			val backlog = endpoint.getMapBacklog(this)

			backlog.clear()
			backlog.add(null to ClearBacklogEntry)

			for ((key, value) in backingMap) {
				val valueCopy = valueCodec.copy(value)

				backlog.add(key to {
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				})
			}
		}

		override val value: MutableMap<K, V> = object : ProxiedMap<K, V>(backingMap) {
			override fun onClear() {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				observingBackingMap?.clear()

				forEachEndpoint {
					it.getMapBacklog(this@Map).also {
						it.clear()
						it.add(null to ClearBacklogEntry)
					}
				}

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueAdded(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val valueCopy = valueCodec.copy(value)

				pushBacklog(key) {
					it.write(MapAction.ADD.ordinal + 1)
					keyCodec.write(it, key)
					valueCodec.write(it, valueCopy)
				}

				observingBackingMap?.put(key, valueCopy)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}

			override fun onValueRemoved(key: K, value: V) {
				if (isRemote) {
					return
				}

				if (endpoints.isEmpty()) {
					return
				}

				val keyCopy = keyCodec.copy(key)

				pushBacklog(key) {
					it.write(MapAction.REMOVE.ordinal + 1)
					keyCodec.write(it, keyCopy)
				}

				observingBackingMap?.remove(key)

				if (!isDirty) {
					notifyEndpoints(this@Map)
					isDirty = true
				}

				hasChanges = true
			}
		}

		override fun write(stream: DataOutputStream, endpoint: Endpoint) {
			stream.write(id)

			sentAllValues = false
			isDirty = false

			val iterator = endpoint.getMapBacklog(this).listIterator()

			for (entry in iterator) {
				entry.second.invoke(stream)
				iterator.remove()
			}

			stream.write(0)
		}

		override fun read(stream: DataInputStream) {
			if (!isRemote) {
				isRemote = true
				clearBacklog()
				observingBackingMap?.clear()
			}

			isDirty = false

			val changeset = LinkedList<MapChangeset<K, V>>()
			var readAction = stream.read() - 1

			while (readAction != -1) {
				if (readAction >= MapActionList.size) {
					throw IndexOutOfBoundsException("Unknown map action with ID $readAction")
				}

				when (MapActionList[readAction]) {
					MapAction.CLEAR -> {
						backingMap.clear()
						changeset.add(ClearMapChangeset)
					}

					MapAction.ADD -> {
						val key = keyCodec.read(stream)
						val value = valueCodec.read(stream)
						backingMap[key] = value
						changeset.add(MapChangeset(MapAction.ADD, key, value))
					}

					MapAction.REMOVE -> {
						val key = keyCodec.read(stream)
						backingMap.remove(key)
						changeset.add(MapChangeset(MapAction.REMOVE, key, null))
					}
				}

				readAction = stream.read() - 1
			}

			if (changeset.size != 0) {
				callback?.invoke(changeset)
			}
		}
	}

	fun invalidate() {
		for (field in fields) {
			field.markDirty()
		}
	}

	fun observe() {
		if (observers.isNotEmpty()) {
			for (field in observers) {
				field.observe()
			}
		}
	}

	/**
	 * [defaultEndpoint]#collectNetworkPayload
	 */
	fun collectNetworkPayload(): FastByteArrayOutputStream? {
		observe()
		return defaultEndpoint.collectNetworkPayload()
	}

	fun applyNetworkPayload(stream: DataInputStream): Int {
		var fieldId = stream.read()
		var i = 0

		while (fieldId != 0) {
			val field = fields.getOrNull(fieldId - 1) ?: throw IndexOutOfBoundsException("Invalid field id $fieldId")
			field.read(stream)
			fieldId = stream.read()
			i++
		}

		if (stream.read() != -1) {
			throw IllegalStateException("Stream wasn't fully drain!")
		}

		return i
	}

	fun applyNetworkPayload(stream: InputStream): Int {
		return applyNetworkPayload(DataInputStream(stream))
	}
}
