package ru.dbotthepony.mc.otm.registry

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import net.minecraft.world.item.DyeColor
import net.minecraftforge.registries.RegistryObject

class LazyList<T> : AbstractList<T> {
	private val getters: Array<() -> T>

	constructor(vararg getters: () -> T) : super() {
		this.getters = Array(getters.size) { getters[it] }
	}

	constructor(getters: List<() -> T>) : super() {
		this.getters = Array(getters.size) { getters[it] }
	}

	override val size: Int
		get() = getters.size

	override fun get(index: Int): T {
		return getters[index].invoke()
	}
}

class ConditionalSet<T> : AbstractSet<T> {
	// method without boxing
	fun interface Condition {
		fun check(): Boolean
	}

	private val getters: Array<Pair<Condition, T>>

	constructor(vararg getters: Pair<Condition, T>) : super() {
		this.getters = Array(getters.size) { getters[it] }
	}

	constructor(getters: List<Pair<Condition, T>>) : super() {
		this.getters = Array(getters.size) { getters[it] }
	}

	override val size: Int get() {
		var i = 0

		for (pair in getters) {
			if (pair.first.check()) {
				i++
			}
		}

		return i
	}

	override fun iterator(): Iterator<T> {
		return object : Iterator<T> {
			val parent = getters.iterator()
			private var pair: Pair<Condition, T>? = null

			private fun search() {
				for (pair in parent) {
					if (pair.first.check()) {
						this.pair = pair
						return
					}
				}

				this.pair = null
			}

			init {
				search()
			}

			override fun hasNext(): Boolean {
				return pair != null
			}

			override fun next(): T {
				val pair = pair ?: throw NoSuchElementException()
				search()
				return pair.second
			}
		}
	}
}

class ConditionalLazySet<T> : AbstractSet<T> {
	// method without boxing
	fun interface Condition {
		fun check(): Boolean
	}

	private val getters: Array<Pair<Condition, () -> T>>

	constructor(vararg getters: Pair<Condition, () -> T>) : super() {
		this.getters = Array(getters.size) { getters[it] }
	}

	constructor(getters: List<Pair<Condition, () -> T>>) : super() {
		this.getters = Array(getters.size) { getters[it] }
	}

	override val size: Int get() {
		var i = 0

		for (pair in getters) {
			if (pair.first.check()) {
				i++
			}
		}

		return i
	}

	override fun iterator(): Iterator<T> {
		return object : Iterator<T> {
			val parent = getters.iterator()
			private var pair: Pair<Condition, () -> T>? = null

			private fun search() {
				for (pair in parent) {
					if (pair.first.check()) {
						this.pair = pair
						return
					}
				}

				this.pair = null
			}

			init {
				search()
			}

			override fun hasNext(): Boolean {
				return pair != null
			}

			override fun next(): T {
				val pair = pair ?: throw NoSuchElementException()
				search()
				return pair.second.invoke()
			}
		}
	}
}

class LazyMap<K, T> : AbstractMap<K, T> {
	override val entries: Set<Map.Entry<K, T>>

	constructor(vararg mValues: Pair<K, () -> T>) : super() {
		entries = ImmutableSet.copyOf(mValues.map { Entry(it.first, it.second) })
	}

	constructor(mValues: Collection<Pair<K, () -> T>>) : super() {
		entries = ImmutableSet.copyOf(mValues.map { Entry(it.first, it.second) })
	}

	constructor(mValues: Map<K, () -> T>) : super() {
		entries = ImmutableSet.copyOf(mValues.map { Entry(it.key, it.value) })
	}

	private inner class Entry(
		override val key: K,
		private val getter: () -> T
	) : Map.Entry<K, T> {
		override val value: T
			get() = getter.invoke()
	}
}

class RegistryObjectList<T>(private vararg val getters: RegistryObject<T>) : AbstractList<T>() {
	constructor(getters: Collection<RegistryObject<T>>) : this(*getters.toTypedArray())

	override val size: Int
		get() = getters.size

	override fun get(index: Int): T {
		return getters[index].get()
	}
}
